@@ -1,6 +1,6 @@
 ;;;; psgml-charent.el
 ;;; Last edited: Mon Nov 28 22:18:09 1994 by lenst@lysistrate (Lennart Staflin)
-;;; $Id: psgml-charent.el,v 1.2 1998/07/10 09:28:24 steveb Exp $
+;;; $Id: psgml-charent.el,v 1.3 2000/03/27 14:10:31 andreasj Exp $
 
 ;; Copyright (C) 1994 Lennart Staflin
 
@@ -99,7 +99,7 @@ Alist with entity name as key and display character as content."
     (save-excursion
       (goto-char (point-min))
       (sgml-with-parser-syntax
-       (while (re-search-forward "&\\(\\w\\(\\w\\|\\s_\\)+\\);?" nil t)
+       (while (re-search-forward "&\\(\\w\\(\\w\\|\\s_\\)*\\);?" nil t)
 	 (setq charent (buffer-substring (match-beginning 1) (match-end 1)))
 	 (if (setq replacement (cdr (assoc charent charent-to-char)))
 	     (replace-match replacement t t)))))))
